;; Generated package description from websocket.el  -*- no-byte-compile: t; lexical-binding:t -*-
(define-package "websocket" "1.16.0.20260228.205745" "Emacs WebSocket client and server" '((cl-lib "0.5")) :commit "2195e1247ecb04c30321702aa5f5618a51c329c5" :authors '(("Andrew Hyatt" . "ahyatt@gmail.com")) :maintainer '("Andrew Hyatt" . "ahyatt@gmail.com") :keywords '("communication" "websocket" "server") :url "https://github.com/ahyatt/emacs-websocket")
